import json

import pandas as pd
import requests

from ..util import util_networkx as nx_util
from ..util import dataframe as df_util

from . import BASE_URL, HEADERS

BASE_URL_NETWORK = BASE_URL + 'networks'
# JSON = 'json'


class CyNetwork(object):

    def __init__(self, suid=None):
        # Validate required argument
        if pd.isnull(suid):
            raise ValueError("SUID is missing.")
        else:
            self.__id = suid

        self.__url = BASE_URL_NETWORK + '/' + str(self.__id) + '/'

    def get_id(self):
        """
        Get session-unique ID of this network

        :return: SUID as integer
        """
        return self.__id

    def to_json(self):
        """
        Return this network in Cytoscape.js format.

        :return: Cytoscape.js Style JSON as dictionary.
        """
        return requests.get(self.__url).json()

    def to_networkx(self):
        """
        Return this network in NetworkX graph object.

        :return: Network as NetworkX graph object
        """
        return nx_util.to_networkx(requests.get(self.__url).json())

    def to_dataframe(self):
        """
        Return this network in pandas DataFrame.

        :return: Network as DataFrame.  This is equivalent to SIF.
        """
        return df_util.to_dataframe(requests.get(self.__url).json())

    def get_nodes(self):
        """
        Get all nodes as a list of SUIDs

        :return:
        """
        return requests.get(self.__url + 'nodes').json()

    def get_edges(self, format='suid'):
        if format is 'suid':
            return requests.get(self.__url + 'edges').json()
        elif format is 'edgelist':
            # TODO: implement this
            pass
        else:
            raise ValueError(format + ' is not supported for edge format.')

    def add_node(self, node_name):
        if node_name is None:
            return None
        return self.add_nodes([node_name])

    def add_nodes(self, node_name_list):
        """
        Add new nodes to the network

        :param node_name_list:
        :return:
        """
        nodes = requests.post(self.__url + 'nodes', data=json.dumps(
            node_name_list), headers=HEADERS).json()
        node_dict = {}
        for node in nodes:
            node_dict[node['name']] = node['SUID']
        return node_dict

    def add_edge(self, source, target, interaction='-', directed=True):
        new_edge = {
            'source': source,
            'target': target,
            'interaction': interaction,
            'directed': directed
        }
        edges = requests.post(self.__url + 'edges', data=json.dumps(
            [new_edge]), headers=HEADERS).json()
        return edges

    def add_edges(self, edge_list):
        new_egdes = []
        for edge_tuple in edge_list:
            new_edge = {
                'source': edge_tuple[0],
                'target': edge_tuple[1],
                'interaction': edge_tuple[2],
            }
            new_egdes.append(new_edge)

        edges = requests.post(self.__url + 'edges', data=json.dumps(
            new_egdes), headers=HEADERS).json()
        df = pd.DataFrame(edges)
        return df.set_index(['SUID'])

    def delete_node(self, id):
        url = self.__url + 'nodes/' + str(id)
        requests.delete(url)

    def delete_edge(self, id):
        url = self.__url + 'edges/' + str(id)
        requests.delete(url)

    def __get_table(self, type, format=None):
        url = self.__url + 'tables/default' + type
        if format is None or format is 'dataframe':
            uri = url + '.tsv'
            return pd.read_csv(uri, sep='\t', index_col=0, header=0)
        elif format is 'csv' or format is 'tsv':
            return requests.get(url + '.' + format).content
        elif format is 'cytoscapejs':
            return requests.get(url).json()['rows']
        else:
            raise ValueError('Unsupported format: ' + format)

    def get_node_table(self, format=None):
        return self.__get_table('node', format)

    def get_edge_table(self, format=None):
        return self.__get_table('edge', format)

    def get_network_table(self, format=None):
        return self.__get_table('network', format)

    def __get_columns(self, type=None):
        url = self.__url + 'tables/default' + type + '/columns'
        df = pd.DataFrame(requests.get(url).json())
        return df.set_index(['name'])

    def get_node_columns(self):
        """
        Get node table column information as DataFrame

        :return: Node column information ad DataFrame
        """
        return self.__get_columns('node')

    def get_edge_columns(self):
        """
        Get edge table column information as DataFrame

        :return: Edge column information ad DataFrame
        """
        return self.__get_columns('edge')

    def get_network_columns(self):
        """
        Get network table column information as DataFrame

        :return: Network column information ad DataFrame
        """
        return self.__get_columns('networks')

    def __get_column(self, type=None, column=None):
        url = self.__url + 'tables/default' + type + '/columns/' + column
        result = requests.get(url).json()
        return pd.Series(result['values'])

    def get_node_column(self, column):
        return self.__get_column('node', column=column)

    def get_edge_column(self, column):
        return self.__get_column('edge', column=column)

    def __get_value(self, type=None, id=None, column=None):
        if column is None and id is not None:
            # Extract a row in table
            url = self.__url + 'tables/default' + type + '/rows/' + str(id)
            return pd.Series(requests.get(url).json())
        elif column is not None and id is not None:
            url = self.__url + 'tables/default' + type + '/rows/' + str(id) + '/' + column
            return requests.get(url).content
        else:
            raise ValueError('ID is required.')

    def get_node_value(self, id, column=None):
        return self.__get_value(type='node', id=id, column=column)

    def get_edge_value(self, id, column=None):
        return self.__get_value(type='edge', id=id, column=column)

    def get_network_value(self, column):
        return self.__get_value(type='network', id=self.__id, column=column)

    def update_node_table(self, df=None, network_key_col='name',
                          data_key_col=None):
        return self.__update_table('node', df=df, network_key_col=network_key_col, data_key_col=data_key_col)

    def __update_table(self, type, df, network_key_col='name',
                       data_key_col=None):

        if data_key_col is None:
            # Use index
            data_key = 'index'
        else:
            data_key = data_key_col

        table = {
            'key': network_key_col,
            'dataKey': data_key
        }
        data = []
        col_names = df.columns.values
        for index, row in df.iterrows():
            entry = {'index': index }
            for col in col_names:
                value = row[col]
                if pd.isnull(value):
                    continue
                else:
                    entry[col] = value
            data.append(entry)

        table['data'] = data
        url = self.__url + 'tables/default' + type
        requests.put(url, data=json.dumps(table), headers=HEADERS)

    def __delete_column(self, type, column):
        url = self.__url + 'tables/default' + type + '/columns/' + column
        requests.delete(url)

    def delete_node_table_column(self, column):
        self.__delete_column('node', column=column)

    def delete_edge_table_column(self, column):
        self.__delete_column('edge', column=column)

    def delete_network_table_column(self, column):
        self.__delete_column('network', column=column)

    def __create_column(self, type, name, data_type, immutable, list):
        url = self.__url + 'tables/default' + type + '/columns'
        new_column = {
            'name': name,
            'type': data_type,
            'immutable': immutable,
            'list': list
        }
        requests.post(url, data=json.dumps(new_column), headers=HEADERS)

    def create_node_column(self, name, data_type='String', is_immutable=False, is_list=False):
        self.__create_column('node', name=name, data_type=data_type, immutable=is_immutable, list=is_list)

    def create_edge_column(self, name, data_type='String', is_immutable=False, is_list=False):
        self.__create_column('edge', name=name, data_type=data_type, immutable=is_immutable, list=is_list)

    def create_network_column(self, name, data_type='String', is_immutable=False, is_list=False):
        self.__create_column('network', name=name, data_type=data_type, immutable=is_immutable, list=is_list)


    # Views
    def get_views(self):
        url = self.__url + 'views'
        return requests.get(url).json()

    def get_first_view(self):
        url = self.__url + 'views/first'
        return requests.get(url).json()

    def get_view(self, view_id):
        url = self.__url + 'views/' + str(view_id)
        return requests.get(url).json()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__id == other.__id
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)



# -*- coding: utf-8 -*-
import os
import requests
import json
import pandas as pd

from . import HEADERS, SUID_LIST
from ..util import cytoscapejs as util

from ..util import util_networkx as nx_util
from ..util import dataframe as df_util

JSON = 'json'

from cynetwork import CyNetwork


class NetworkClient(object):

    def __init__(self, url):
        self.__url = url + 'networks'

    def create_from(self, locations=None, collection=None):
        if locations is None:
            raise ValueError('Locations parameter is required.')

        input_type = type(locations)
        if input_type is list or input_type is tuple or input_type is set:
            location_list = []
            for loc in locations:
                if not str(loc).startswith('http'):
                    location_list.append(self.__to_file_url(loc))
                else:
                    location_list.append(loc)
        else:
            if not str(locations).startswith('http'):
                location_list = [self.__to_file_url(locations)]
            else:
                location_list = [locations]

        if collection is None:
            collection_name = 'Created from resources'
        else:
            collection_name = collection

        parameters = {
            'collection': collection_name,
            'source': 'url'
        }

        res = requests.post(self.__url, data=json.dumps(location_list),
                             params=parameters,
                             headers=HEADERS).json()

        if len(res) == 1:
            network_ids = res[0]['networkSUID']
            if len(network_ids) == 1:
                return CyNetwork(network_ids[0])
            else:
                return [CyNetwork(suid) for suid in network_ids]
        else:
            result_dict = {entry['source']: CyNetwork(entry['networkSUID'])
                           for entry in res}
            return pd.Series(result_dict)

    def __to_file_url(self, file_name):
        local_file = os.path.abspath(file_name)
        return 'file:///' + local_file

    def create(self, suid=None, name=None, collection=None,
               data=None):
        if suid is not None:
            # fetch existing network
            res = requests.get(self.__url)
            existing_networks = res.json()
            if suid in existing_networks:
                network_id = suid
            else:
                raise ValueError('No such network')
        else:
            if data is None:
                network_data = util.get_empty_network()
            else:
                network_data = data

            if name is not None:
                network_data['data']['name'] = name

            if collection is None:
                network_collection = 'From cyREST'
            else:
                network_collection = collection

            res = requests.post(self.__url + '?collection=' + network_collection, data=json.dumps(network_data), headers=HEADERS)
            result = res.json()
            network_id = result['networkSUID']

        return CyNetwork(network_id)

    def create_from_networkx(self, network, name=None, collection=None):
        if network is None:
            raise ValueError('NetworkX graph object is required.')

        cyjs = nx_util.from_networkx(network)
        return self.create(name=name, collection=collection, data=cyjs)

    def create_from_dataframe(self, dataframe, source_col='source',
                              target_col='target', interaction_col='interaction',
                              name='Created from DataFrame', collection=None):
        if dataframe is None:
            raise ValueError('DataFrame object is required.')

        # Convert from DataFrame to Cytoscape.js JSON
        cyjs = df_util.from_dataframe(dataframe, source_col=source_col,
                              target_col=target_col, interaction_col=interaction_col,
                              name=name)
        return self.create(collection=collection, data=cyjs)

    def get_all(self, format=SUID_LIST):
        if format is SUID_LIST:
            result = requests.get(self.__url)
        elif format is JSON:
            result = requests.get(self.__url + '.json')
        else:
            raise ValueError('Unsupported format type: ' + format)

        return result.json()

    def get(self, id):
        return requests.get(self.__url + '/' + str(id) + '.json').json()

    def delete_all(self):
        requests.delete(self.__url)

    def delete(self, cynetwork):
        id = cynetwork.get_id()
        requests.delete(self.__url + '/' + str(id))
        del cynetwork




# -*- coding: utf-8 -*-

import unittest

from py2cytoscape.data.style import Style
from py2cytoscape.data.cyrest_client import CyRestClient

import pandas as pd
import json

STYLE_NAME = 'style1'


class StyleTests(unittest.TestCase):

    def setUp(self):
        self.client = CyRestClient()
        self.client.style.delete_all()
        styles = self.client.style.get_all()
        self.assertEqual(1, len(styles))

        self.style = self.client.style.create(STYLE_NAME)
        self.assertEqual(STYLE_NAME, self.style.get_name())

    def test_get_style(self):
        style_as_cy3_json = self.client.style.get(STYLE_NAME)
        style_as_cyjs_json = self.client.style.get(STYLE_NAME,
                                                   data_format='cytoscapejs')
        print(style_as_cy3_json)
        print(json.dumps(style_as_cyjs_json,indent=4))

    def test_mappings(self):
        print('\n---------- Mappings tests start -----------\n')

        mappings = self.style.get_mappings()
        self.assertIsNotNone(mappings)
        self.assertEqual(list, type(mappings))
        self.assertEqual(0, len(mappings))

        print('\n---------- Mapping tests finished! -----------\n')

    def test_defaults(self):
        print('\n---------- Defaults tests start -----------\n')

        default_style = self.client.style.create('default')
        defs = default_style.get_defaults()
        print(defs)
        self.assertIsNotNone(defs)
        self.assertEqual(pd.Series, type(defs))
        self.assertEqual(103, len(defs))

        defs2 = self.style.get_defaults()
        print(defs2)
        self.assertIsNotNone(defs2)
        self.assertEqual(pd.Series, type(defs2))
        self.assertEqual(103, len(defs2))

        print('\n---------- Defaults tests finished! -----------\n')

    def test_get_defaults(self):
        print('\n---------- GET Defaults tests start -----------\n')

        default_style = self.client.style.create('def2')
        n_label = default_style.get_default('NODE_LABEL')
        n_size = default_style.get_default('NODE_SIZE')
        n_shape = default_style.get_default('NODE_SHAPE')
        e_width = default_style.get_default('EDGE_WIDTH')

        print(n_label)
        # These should be same as VP defaults.
        self.assertEqual(pd.Series, type(n_label))
        self.assertEqual('', n_label.ix['NODE_LABEL'])
        self.assertEqual(50, n_size.ix['NODE_SIZE'])
        self.assertEqual('ELLIPSE', n_shape.ix['NODE_SHAPE'])
        self.assertEqual(1, e_width.ix['EDGE_WIDTH'])

    def test_update_defaults(self):
        print('\n---------- Add Defaults tests start -----------\n')

        default_style = self.client.style.create('default')

        new_defaults = {
            # Node defaults
            'NODE_FILL_COLOR': '#eeeeff',
            'NODE_SIZE': 20,
            'NODE_BORDER_WIDTH': 0,
            'NODE_TRANSPARENCY': 120,
            'NODE_LABEL_COLOR': 'white',

            # Edge defaults
            'EDGE_WIDTH': 9,
            'EDGE_STROKE_UNSELECTED_PAINT': '#aaaaaa',
            'EDGE_LINE_TYPE': 'LONG_DASH',
            'EDGE_TRANSPARENCY': 120,

            # Network defaults
            'NETWORK_BACKGROUND_PAINT': 'black'
        }

        default_style.update_defaults(new_defaults)

        defs = default_style.get_defaults()
        self.assertEqual(20, defs.ix['NODE_SIZE'])
        self.assertEqual('#eeeeff'.upper(), defs.ix['NODE_FILL_COLOR'])
        self.assertEqual(0, defs.ix['NODE_BORDER_WIDTH'])
        self.assertEqual(120, defs.ix['NODE_TRANSPARENCY'])
        self.assertEqual('#FFFFFF', defs.ix['NODE_LABEL_COLOR'])

        self.assertEqual(9, defs.ix['EDGE_WIDTH'])
        self.assertEqual('#aaaaaa'.upper(), defs.ix['EDGE_STROKE_UNSELECTED_PAINT'])
        self.assertEqual('LONG_DASH', defs.ix['EDGE_LINE_TYPE'])
        self.assertEqual(120, defs.ix['EDGE_TRANSPARENCY'])

        # Network defaults
        self.assertEqual('#000000', defs.ix['NETWORK_BACKGROUND_PAINT'])

        print('\n---------- Defaults tests finished! -----------\n')

    def test_passthrough_mappings(self):
        print('---------- Passthrough mmapping tests -----------\n')

        self.style.create_passthrough_mapping(column='name',
                                              col_type='String', vp='NODE_LABEL')
        mappings = self.style.get_mappings()
        self.assertEqual(1, len(mappings))
        pt = self.style.get_mapping('NODE_LABEL')
        print(pt)
        self.assertEqual('passthrough', pt['mappingType'])
        self.assertEqual('NODE_LABEL', pt['visualProperty'])
        self.assertEqual('String', pt['mappingColumnType'])
        self.assertEqual('name', pt['mappingColumn'])

        self.style.delete_mapping(vp='NODE_LABEL')
        mappings = self.style.get_mappings()
        self.assertEqual(0, len(mappings))

        print('---------- Passthrough mapping tests finished! -----------\n')

    def test_discrete_mappings(self):
        print('---------- Discrete mapping tests -----------\n')

        vp = 'EDGE_LINE_TYPE'
        col = 'interaction'

        try:
            self.style.create_discrete_mapping(column=col,
                                           col_type='String',
                                           vp=vp)
        except ValueError as e:
            print(e)
        else:
            raise RuntimeError('This should not happen!!')

        mapping = {
            'pp': 'SOLID',
            'pd': 'LONG_DASH'
        }

        self.style.create_discrete_mapping(column=col,
                                               col_type='String',
                                               vp=vp, mappings=mapping)
        dm = self.style.get_mapping(vp)
        print(dm)
        self.assertEqual('discrete', dm['mappingType'])
        self.assertEqual(vp, dm['visualProperty'])
        self.assertEqual('String', dm['mappingColumnType'])
        self.assertEqual(col, dm['mappingColumn'])
        self.assertIsNotNone(dm['map'])

        dm_map = dm['map']
        for entry in dm_map:
            self.assertTrue(entry['key'] in mapping.keys())
            self.assertEqual(entry['value'], mapping[entry['key']])

        print('---------- Discrete mapping tests finished! -----------\n')



import requests
from . import HEADERS, SUID_LIST


class LayoutClient(object):

    def __init__(self, url):
        self.__url = url + 'apply/layouts'

    def get_all(self):
        return requests.get(self.__url).json()

    def apply(self, name='force-directed', network=None):
        if network is None:
            raise ValueError('Target network is required')

        url = self.__url + '/' + name + '/' + str(network.get_id())
        requests.get(url)

    def create_discrete_mapping(self, column=None, visual_property=None, mapping=None):
        pass



# -*- coding: utf-8 -*-

import unittest

import networkx as nx

from py2cytoscape.data.cyrest_client import CyRestClient
import py2cytoscape.util.dataframe as df_util
import json


def pp(dict_data):
    print(json.dumps(dict_data, indent=4))


class CyRestClientTests(unittest.TestCase):

    def setUp(self):
        self.client = CyRestClient()
        # cleanup
        self.client.network.delete_all()

    def test_cyrest_client(self):
        print('\n---------- Client status tests start -----------\n')
        # Make sure CyREST server is running
        status = self.client.status()
        self.assertIsNotNone(status)
        pp(status)
        self.assertEqual('v1', status['apiVersion'])

        print('\n---------- Client status tests finished! -----------\n')

    def test_create_network(self):
        print('\n---------- Create network Tests Start -----------\n')
        # Create empty network
        num_networks = 5
        for i in range(num_networks):
            self.client.network.create()

        networks = self.client.network.get_all()
        self.assertIsNotNone(networks)
        self.assertEqual(num_networks, len(networks))

    def test_create_from(self):
        print('\n---------- Loading network tests start -----------\n')
        # Create from URL
        locations = [
            'http://chianti.ucsd.edu/cytoscape-data/galFiltered.sif',
            'http://www.ebi.ac.uk/Tools/webservices/psicquic/intact/'
            + 'webservices/current/search/interactor/brca2_human?format=xml25'

        ]

        networks = self.client.network.create_from(locations)
        print(networks)
        self.assertIsNotNone(networks)
        print('---------- Loading networks done! -----------\n')

    def test_network_api(self):
        print('\n---------- Network API Tests Start -----------\n')
        # Create empty network
        for i in range(5):
            self.client.network.create()

        networks = self.client.network.get_all(format='json')
        self.assertIsNotNone(networks)
        pp(networks)

    def test_cynetwork(self):
        print('\n---------- CyNetwork Tests Start -----------\n')
        network = self.client.network.create()
        self.assertIsNotNone(network)
        nodes = network.get_nodes()
        pp(nodes)
        new_nodes = ['a', 'b', 'c']
        nd = network.add_nodes(new_nodes)

        self.assertIsNotNone(nd)
        pp(nd)

        edge = network.add_edge(nd['a'], nd['b'])
        self.assertIsNotNone(edge)
        pp(edge)

        new_edges = ((nd['a'], nd['b'], 'i1'), (nd['a'], nd['c'], 'i1'))
        new_edges_result = network.add_edges(new_edges)
        print(new_edges_result)
        self.assertEqual(2, len(new_edges_result))

        node_table = network.get_node_table()
        print(node_table)
        node_table = network.get_node_table(format='tsv')
        print(node_table)
        node_table = network.get_node_table(format='csv')
        print(node_table)

        edge_table = network.get_edge_table()
        print(edge_table)

        print('\n---------- CyNetwork Tests Finished! -----------\n')

    def test_convert(self):
        print('\n---------- DataFrame Conversion Tests Start -----------\n')
        import os
        import pandas as pd
        dir_name = os.path.dirname(os.path.realpath(__file__))
        df = pd.read_csv(
            dir_name + '/data/galFiltered.sif',
            names=['source', 'interaction', 'target'], sep=' ')
        print(df.head(3))
        net = df_util.from_dataframe(df)

        network = self.client.network.create(
            data=net, name='Created from DataFrame')
        dir_name = os.path.dirname(os.path.realpath(__file__))
        file_name = dir_name + '/data/galFiltered.nodeAttrTable.txt'
        data_table = pd.read_csv(file_name, sep='\t')
        network.update_node_table(df=data_table, data_key_col='ID')
        print('\n---------- DataFrame Conversion Tests Finished! -----------\n')

    def test_delete_network(self):
        network = self.client.network.create()
        suids = self.client.network.get_all()
        self.assertEqual(1, len(suids))
        self.client.network.delete(network)
        suids = self.client.network.get_all()
        self.assertEqual(0, len(suids))

    def test_create_from_networkx(self):
        networkx1 = nx.scale_free_graph(100)

        network = self.client.network.create_from_networkx(networkx1)
        self.assertIsNotNone(network)
